/*

net.c - Networking

*/

/*...sincludes:0:*/
#include <stdio.h>
#include <string.h>
#include "interrupts.h"
#include "list.h"
#include "fiber.h"
#include "cpmio.h"
#include "wiz.h"
#include "net.h"
#include "log.h"

/*...vinterrupts\46\h:0:*/
/*...vlist\46\h:0:*/
/*...vfiber\46\h:0:*/
/*...vsem\46\h:0:*/
/*...vcpmio\46\h:0:*/
/*...vwiz\46\h:0:*/
/*...vnet\46\h:0:*/
/*...e*/

/*...svars:0:*/
static byte ip_subnet[]  = { 255,255,255,0 };
static byte ip[]         = { 192,168,1,123 }; /* chip errata, cannot be 0.0.0.0 */
static byte ip_gateway[] = { 192,168,1,1 };

// https://en.wikipedia.org/wiki/MAC_address
// http://standards-oui.ieee.org/oui/oui.txt
// 00-08-DC-xx-xx-xx is Wiznet
// x2-xx-xx-xx-xx-xx is locally administered unicast
static byte mac[] = { 0x00,0x08,0xdc, 'N','F','X' };

static FCB net_fcb;
static byte net_dma[128];
static byte *net_pch = NULL;

#define	N_SOCKETS 4
static SOCKET net_sockets[N_SOCKETS];
static LIST net_sockets_free;
static LIST net_sockets_waiters;

static word net_ephemeral = 0; 
/*...e*/

/*...snet_fatal:0:*/
static const char *psStatus[] = {
    " SS_FREE\r\n$",
    " SS_ALLOCATED\r\n$",
    " SS_CONNECT\r\n$",
    " SS_LISTEN\r\n$",
    " SS_ESTABLISHED\r\n$",
    " SS_SEND\r\n$",
    " SS_RECV\r\n$",
    " SS_ERROR\r\n$",
    " SS_DISCONNECT\r\n$"
    };

static void net_fatal(SOCKET *socket, char *error)
	{
    if ( socket != NULL )
        {
        char *ps = strchr (error, '?');
        if ( ps != NULL )
            *ps = '0' + ((socket->sra - W5100_ADDR_SOCKET) >> 8);
        }
	interrupts_disable();
	log_printf(error);
    log_printf ("Socket $");
    log_uint ((socket->sra - W5100_ADDR_SOCKET) >> 8);
    log_printf (psStatus[socket->state]);
    log_close();
    bdos_exit();
	}
/*...e*/

static char *byte_dec(char *p, byte b)
    {
    char c1 = '0';
    char c2 = '0';
    while (b >= 100)
        {
        ++c1;
        b -= 100;
        }
    if (c1 > '0') *(p++) = c1;
    while (b >= 10)
        {
        ++c2;
        b -= 10;
        }
    if ((c1 > '0') || (c2 > '0'))  *(p++) = c2;
    *(p++) = b + '0';
    return p;
    }

static void cfg_err(const char *err, byte c, byte n)
    {
    char info[] = "'?' (??h) at ?  \r\n$";
    byte h;
    log_printf(err);
    info[1] = c;
    h = c >> 4;
    info[5] = ( h < 10 ) ? h + '0' : h - 10 + 'A';
    h = c & 0x0F;
    info[6] = ( h < 10 ) ? h + '0' : h - 10 + 'A';
    byte_dec(info + 13, n);
    net_fatal(NULL, info);
    }

static byte net_ch(void)
    {
    byte b;
    if ( net_pch == NULL ) return 0x1A;
    if ( net_pch >= net_dma + 128 )
        {
        if ( bdos_read_file (&net_fcb) != 0 )
            {
            net_pch = NULL;
            return 0x1A;
            }
        net_pch = net_dma;
        }
    b = *(net_pch++);
    if ( b == 0x1A ) net_pch = NULL;
    return b;
    }

static void net_parse_ip (byte *ip)
    {
    int i = 0;
    word b = 0;
    byte c = net_ch();
    byte n = 1;
    byte d = 0;
    while ((c == ' ') || (c == '\t'))
        {
        if ( c == 0x1A ) return;
        c = net_ch();
        ++n;
        }
    while (TRUE)
        {
        if (( c >= '0' ) && ( c <= '9' ))
            {
            b = 10 * b + ( c - '0' );
            ++d;
            if ( b > 255 ) cfg_err("Byte overflow: $", c, n);
            }
        else if ( c == '.' )
            {
            if ( i > 2 ) cfg_err("Too many parts: $", c, n);
            if ( d == 0 ) cfg_err("Missing digits: $", c, n);
            ip[i] = b;
            b = 0;
            d = 0;
            ++i;
            }
        else if ((c == ' ') || (c == '\t') || (c == '\r') || (c == '\n') || (c == 0x1A))
            {
            break;
            }
        else
            {
            cfg_err("Invalid character: $", c, n);
            }
        c = net_ch();
        ++n;
        }
    if ( i == 3 ) ip[i] = b;
    else cfg_err("Missing parts: $", c, n);
    while ((c != '\n') && (c != 0x1A))
        {
        if ((c != ' ') && (c != '\t') && (c != '\r')) cfg_err("Invalid character: $", c, n);
        c = net_ch();
        ++n;
        }
    }

static void net_parse_mac (byte *mac)
    {
    word b = 0;
    byte c = net_ch();
    int i = 0;
    byte n = 1;
    byte d = 0;
    while ((c == ' ') || (c == '\t'))
        {
        if ( c == 0x1A ) return;
        c = net_ch();
        ++n;
        }
    while (TRUE)
        {
        if (( c >= '0' ) && ( c <= '9' ))
            {
            b = 16 * b + ( c - '0' );
            ++d;
            }
        else if (( c >= 'A' ) && ( c <= 'F' ))
            {
            b = 16 * b + ( c - 'A' + 10 );
            ++d;
            }
        else if (( c >= 'a' ) && ( c <= 'f' ))
            {
            b = 16 * b + ( c - 'a' + 10 );
            ++d;
            }
        else if (( c == ':' ) || (c == '-'))
            {
            if ( i > 4 ) cfg_err("Too many parts: $", c, n);
            if ( d == 0 ) cfg_err("Missing digits: $", c, n);
            mac[i] = b;
            b = 0;
            d = 0;
            ++i;
            }
        else if ((c == ' ') || (c == '\t') || (c == '\r') || (c == '\n') || (c == 0x1A))
            {
            break;
            }
        else
            {
            cfg_err("Invalid character: $", c, n);
            }
        if ( b > 255 ) cfg_err("Byte overflow: $", c, n);
        c = net_ch();
        ++n;
        }
    if ( i == 5 ) mac[i] = b;
    else cfg_err("Missing parts: $", c, n);
    while ((c != '\n') && (c != 0x1A))
        {
        if ((c != ' ') && (c != '\t') && (c != '\r')) cfg_err("Invalid character: $", c, n);
        c = net_ch();
        ++n;
        }
    }

static void show_ip(byte *ip)
    {
    char addr[20];
    char *p = addr;
    p = byte_dec(p, ip[0]);
    *(p++) = '.';
    p = byte_dec(p, ip[1]);
    *(p++) = '.';
    p = byte_dec(p, ip[2]);
    *(p++) = '.';
    p = byte_dec(p, ip[3]);
    *(p++) = '\r';
    *(p++) = '\n';
    *(p++) = '$';
    log_printf(addr);
    }

static void show_mac(byte *mac)
    {
    char addr[24];
    char *p = addr;
    byte b;
    int i;
    for (i = 0; i < 6; ++i)
        {
        b = mac[i] >> 4;
        *(p++) = ( b < 10 ) ? b + '0' : b - 10 + 'A';
        b = mac[i] & 0x0F;
        *(p++) = ( b < 10 ) ? b + '0' : b - 10 + 'A';
        *(p++) = ':';
        }
    --p;
    *(p++) = '\r';
    *(p++) = '\n';
    *(p++) = '$';
    log_printf(addr);
    }

static void net_config (void)
    {
    net_fcb.drive = 0;
    net_fcb.filename[0] = 'N';
    net_fcb.filename[1] = 'F';
    net_fcb.filename[2] = 'X';
    net_fcb.filename[3] = ' ';
    net_fcb.filename[4] = ' ';
    net_fcb.filename[5] = ' ';
    net_fcb.filename[6] = ' ';
    net_fcb.filename[7] = ' ';
    net_fcb.type[0] = 'C';
    net_fcb.type[1] = 'F';
    net_fcb.type[2] = 'G';
    net_fcb.ex = 0;
    net_fcb.s1 = 0;
    net_fcb.s2 = 0;
    net_fcb.cr = 0;
    if ( bdos_open_file (&net_fcb) != 255 )
        {
        net_fcb.ex = 0;
        net_fcb.cr = 0;
        bdos_set_dma_address (net_dma);
        net_pch = net_dma + 128;
        }
    log_printf("IP address:  $");
    if ( net_pch != NULL ) net_parse_ip (ip);
    show_ip(ip);
    log_printf("Net mask:    $");
    if ( net_pch != NULL ) net_parse_ip (ip_subnet);
    show_ip(ip_subnet);
    log_printf("Gateway:     $");
    if ( net_pch != NULL ) net_parse_ip (ip_gateway);
    show_ip(ip_gateway);
    log_printf("MAC address: $");
    if ( net_pch != NULL ) net_parse_mac (mac);
    show_mac(mac);
    }

/*...sdiags:0:*/
static char nibble[] = "_$";
static void outnibble(byte n)
	{
	nibble[0] = ( n < 10 ) ? ( n+'0' ) : ( -10+'A' );
	log_printf(nibble);
	}
static void outbyte(byte b)
	{
	outnibble(b>>4);
	outnibble(b   );
	}
/*...e*/

/*...snet_alloc:0:*/
void net_alloc(SOCKET_ALLOC *sockalloc, NF nf, void *nfp)
	{
	SOCKET *socket = list_remove_first(&net_sockets_free);
	if ( socket != NULL )
		{
		socket->state = SS_ALLOCATED;
		sockalloc->socket = socket;
		nf(nfp);
		}
	else
		{
		sockalloc->fiber.nf  = nf;
		sockalloc->fiber.nfp = nfp;
		list_add_last(&net_sockets_waiters, sockalloc);
		}
	}
/*...e*/
/*...snet_free:0:*/
void net_free(SOCKET *socket)
	{
	SOCKET_ALLOC *sockalloc = list_remove_first(&net_sockets_waiters);
	if ( socket->state != SS_ALLOCATED )
		net_fatal(socket, "net_free: socket ? not SS_ALLOCATED\r\n$");
	if ( sockalloc != NULL )
		{
		sockalloc->socket = socket;
		fiber_enqueue((FIBER *) sockalloc);
		}
	else
		{
		socket->state = SS_FREE;
		list_add_first(&net_sockets_free, socket);
		}
	}
/*...e*/

/*...snet_next_ephemeral_port:0:*/
word net_next_ephemeral_port(void)
	{
	return 0x8000 + (net_ephemeral++) % 0x7fff;
	}
/*...e*/
/*...snet_init_socket_as_tcp:0:*/
static void net_init_socket_as_tcp(word sra, word port)
	{
	for ( ;; )
		{
		wiz_write(sra+W5100_OFFSET_SOCKET_MR, W5100_SOCKET_MR_TCP);
		wiz_write_word(sra+W5100_OFFSET_SOCKET_PORT0, port);
		wiz_write(sra+W5100_OFFSET_SOCKET_CR, W5100_SOCKET_CR_OPEN);
		if ( wiz_read(sra+W5100_OFFSET_SOCKET_SR) == W5100_SOCKET_SR_SOCK_INIT )
			break;
		wiz_write(sra+W5100_OFFSET_SOCKET_CR, W5100_SOCKET_CR_CLOSE);
		}
	}
/*...e*/
/*...snet_complete_op:0:*/
static void net_complete_op(SOCKET *socket)
	{
	// (socket->fiber_op.nf)(socket->fiber_op.nfp);
#ifdef DEBUG
    log_printf ("net_complete_op: Socket $");
    log_uint ((socket->sra - W5100_ADDR_SOCKET) >> 8);
    log_printf (" enqueue $");
    log_word ((word) socket->fiber_op.nf);
    log_printf ("\r\n$");
#endif
    fiber_enqueue (&socket->fiber_op);
	}
/*...e*/

/*...snet_connect:0:*/
static void net_connect_ok(SOCKET *socket)
	{
	socket->state = SS_ESTABLISHED;
	net_complete_op(socket);
	}

static void net_connect_error(SOCKET *socket)
	{
	word sra = socket->sra;
	wiz_write(sra+W5100_OFFSET_SOCKET_CR, W5100_SOCKET_CR_CLOSE);
	socket->state = SS_ALLOCATED;
	net_complete_op(socket);
	}

void net_connect(SOCKET *socket, const byte *ip, word port, NF nf, void *nfp)
	{
	word sra = socket->sra;
	if ( socket->state != SS_ALLOCATED )
		net_fatal(socket, "net_connect: socket ? not SS_ALLOCATED\r\n$");
    if ( socket->fiber_op.queued ) fiber_dequeue (&socket->fiber_op);
    // net_fatal(socket, "net_connect: socket ? fiber already queued\r\n$");
#ifdef DEBUG
    log_printf ("net_connect: Socket $");
    log_uint ((socket->sra - W5100_ADDR_SOCKET) >> 8);
//     log_printf (" callback $");
//     log_word ((word) nf);
    log_printf ("\r\n$");
#endif
	socket->fiber_op.nf  = nf;
	socket->fiber_op.nfp = nfp;
	socket->consumed = 0;
	net_init_socket_as_tcp(sra, net_next_ephemeral_port());
	wiz_write_n(sra+W5100_OFFSET_SOCKET_DIPR0, ip, 4);
	wiz_write_word(sra+W5100_OFFSET_SOCKET_DPORT0, port);
	wiz_write(sra+W5100_OFFSET_SOCKET_CR, W5100_SOCKET_CR_CONNECT);
	socket->state = SS_CONNECT;
	}
/*...e*/
/*...snet_listen:0:*/
static void net_listen_ok(SOCKET *socket)
	{
#ifdef DEBUG
    log_printf ("net_listen_ok: Socket $");
    log_uint ((socket->sra - W5100_ADDR_SOCKET) >> 8);
    log_printf ("\r\n$");
#endif
	socket->state = SS_ESTABLISHED;
	net_complete_op(socket);
	}

void net_listen(SOCKET *socket, word port, NF nf, void *nfp)
	{
	word sra = socket->sra;
	if ( socket->state != SS_ALLOCATED )
		net_fatal(socket, "net_listen: socket ? not SS_ALLOCATED\r\n$");
    if ( socket->fiber_op.queued ) fiber_dequeue (&socket->fiber_op);
    // net_fatal(socket, "net_listen: socket ? fiber already queued\r\n$");
#ifdef DEBUG
    log_printf ("net_listen: Socket $");
    log_uint ((socket->sra - W5100_ADDR_SOCKET) >> 8);
    log_printf (" port $");
    log_word (port);
    log_printf ("\r\n$");
#endif
	socket->fiber_op.nf  = nf;
	socket->fiber_op.nfp = nfp;
	socket->consumed = 0;
	net_init_socket_as_tcp(sra, port);
	wiz_write(sra+W5100_OFFSET_SOCKET_CR, W5100_SOCKET_CR_LISTEN);
	socket->state = SS_LISTEN;
	}
/*...e*/
/*...snet_disconnect:0:*/
static void net_disconnect_ok(SOCKET *socket)
	{
	word sra = socket->sra;
	wiz_write(sra+W5100_OFFSET_SOCKET_CR, W5100_SOCKET_CR_CLOSE);
	socket->state = SS_ALLOCATED;
	net_complete_op(socket);
	}

void net_disconnect(SOCKET *socket, NF nf, void *nfp)
	{
	word sra = socket->sra;
	if ( socket->state != SS_ESTABLISHED && socket->state != SS_ERROR )
		net_fatal(socket, "net_disconnect: socket ? not SS_ESTABLISHED or SS_ERROR\r\n$");
    if ( socket->fiber_op.queued ) fiber_dequeue (&socket->fiber_op);
    // net_fatal(socket, "net_disconnect: socket ? fiber already queued\r\n$");
#ifdef DEBUG
    log_printf ("net_disconnect: Socket $");
    log_uint ((socket->sra - W5100_ADDR_SOCKET) >> 8);
//     log_printf (" callback $");
//     log_word ((word) nf);
    log_printf ("\r\n$");
#endif
	socket->fiber_op.nf  = nf;
	socket->fiber_op.nfp = nfp;
	wiz_write(sra+W5100_OFFSET_SOCKET_CR, W5100_SOCKET_CR_DISCON);
	socket->state = SS_DISCONNECT;
	}
/*...e*/

/*...snet_send:0:*/
static void net_send_chunk_error(SOCKET *socket)
	{
	socket->state = SS_ERROR;
	net_complete_op(socket);
	}

static void show_reg(const char *reg, word val)
    {
    static char hex[] = "????\r\n$";
    byte h;
    log_printf (reg);
    h = val >> 12;
    hex[0] = ( h < 10 ) ? h + '0' : h - 10 + 'A';
    h = ( val >> 8 ) & 0x0F;
    hex[1] = ( h < 10 ) ? h + '0' : h - 10 + 'A';
    h = ( val >> 4 ) & 0x0F;
    hex[2] = ( h < 10 ) ? h + '0' : h - 10 + 'A';
    h = val & 0x0F;
    hex[3] = ( h < 10 ) ? h + '0' : h - 10 + 'A';
    log_printf (hex);
    }

static void net_send_chunk(SOCKET *socket)
	{
#ifdef DEBUG
    log_printf ("net_send_chunk: socket $");
    log_uint (((socket->sra - W5100_ADDR_SOCKET) >> 8));
    log_printf (" nf = $");
    log_word ((word) socket->fiber_op.nf);
#endif
	if ( socket->actual == socket->length )
		{
#ifdef DEBUG
        log_printf (" - Completed\r\n$");
#endif
		socket->state = SS_ESTABLISHED;
		net_complete_op(socket);
		}
	else
		{
#ifdef DEBUG
        log_printf ("\r\n$");
#endif
		word sra = socket->sra;
		word wr = wiz_read_word(sra+W5100_OFFSET_SOCKET_TX_WR0);
		word wr_modulo = wr & 2047;
		word thisgo = socket->length - socket->actual;
        word fsr = wiz_read_word(sra+W5100_OFFSET_SOCKET_TX_FSR0);
		if ( thisgo > 2048 ) thisgo = 2048;
        // Diagnostic code
        /*
        word rd = wiz_read_word(sra+W5100_OFFSET_SOCKET_TX_RD0);
        show_reg("s_mr = $", wiz_read(sra+W5100_OFFSET_SOCKET_MR));
        show_reg("s_cr = $", wiz_read(sra+W5100_OFFSET_SOCKET_CR));
        show_reg("ir_last = $", ir_last);
        show_reg("s_sr = $", wiz_read(sra+W5100_OFFSET_SOCKET_SR));
        show_reg("s_tx_fsr = $", fsr);
        show_reg("s_tx_rd = $", rd);
        show_reg("s_tx_wr = $", wr);
        show_reg("wr_modulo = $", wr_modulo);
        show_reg("thisgo = $", thisgo);
        */
		if ( fsr < thisgo )
            {
            if ( ! socket->fiber_int.queued )
                {
                // Insufficient free space - try again later
                socket->fiber_int.nf     = (NF)net_send_chunk;
                socket->fiber_int.nfp    = socket;
                // socket->fiber_int_queued = TRUE;
                fiber_enqueue(&(socket->fiber_int));
                log_printf ("Retrying\r\n$");
                }
            else
                {
                net_fatal(socket, "net_send_chunk: socket ? no free space\r\n$");
                }
            }
		if ( wr_modulo + thisgo <= 2048 )
			wiz_write_n(socket->tx_buf+wr_modulo, socket->send_ptr, thisgo);
		else
			{
			word thisgo1 = 2048-wr_modulo;
			word thisgo2 = thisgo-thisgo1;
			wiz_write_n(socket->tx_buf+wr_modulo, socket->send_ptr        , thisgo1);
			wiz_write_n(socket->tx_buf          , socket->send_ptr+thisgo1, thisgo2); 
			}
		wiz_write_word(sra+W5100_OFFSET_SOCKET_TX_WR0, wr+thisgo);
		socket->thisgo = thisgo;
		wiz_write(sra+W5100_OFFSET_SOCKET_CR, W5100_SOCKET_CR_SEND);
		}
	}

static void net_send_chunk_ok(SOCKET *socket)
	{
	socket->send_ptr += socket->thisgo;
	socket->actual   += socket->thisgo;
	net_send_chunk(socket);
	}

void net_send(SOCKET *socket, const byte *ptr, size_t length, NF nf, void *nfp)
	{
	word sra = socket->sra;
	if (( socket->state != SS_ESTABLISHED ) && ( socket->state != SS_RECV ))
		net_fatal(socket, "net_send: socket ? not SS_ESTABLISHED\r\n$");
    if ( socket->fiber_op.queued ) fiber_dequeue (&socket->fiber_op);
    // net_fatal(socket, "net_send: socket ? fiber already queued\r\n$");
#ifdef DEBUG
    log_printf ("net_send: Socket $");
    log_uint ((socket->sra - W5100_ADDR_SOCKET) >> 8);
    log_printf (" callback $");
    log_word ((word) nf);
    log_printf ("\r\n$");
#endif
	socket->fiber_op.nf  = nf;
	socket->fiber_op.nfp = nfp;
	socket->send_ptr     = ptr;
	socket->length       = length;
	socket->actual       = 0;
	socket->state        = SS_SEND;
	net_send_chunk(socket);
	}
/*...e*/
/*...snet_recv:0:*/
static void net_recv_error(SOCKET *socket)
	{
	socket->state = SS_ERROR;
	net_complete_op(socket);
	}

static void net_recv_consume(SOCKET *socket, word available)
	{
	word sra = socket->sra;
	word rd = wiz_read_word(sra+W5100_OFFSET_SOCKET_RX_RD0) + socket->consumed;
	word rd_modulo = rd & 2047;
	word thisgo = socket->length;
#ifdef DEBUG
    log_printf("net_recv_consume: $");
    show_reg("socket = ", (word) socket);
    show_reg("consumed = $", socket->consumed);
    show_reg("available = $", available);
#endif
	if ( thisgo > available )
		thisgo = available;
#ifdef DEBUG
    show_reg("thisgo = $", thisgo);
#endif
	if ( rd_modulo + thisgo <= 2048 )
		wiz_read_n(socket->rx_buf+rd_modulo, socket->recv_ptr, thisgo);
	else
		{
		word thisgo1 = 2048-rd_modulo;
		word thisgo2 = thisgo-thisgo1;
        show_reg ("thisgo1 = ", thisgo1);
        show_reg ("thisgo2 = ", thisgo2);
		wiz_read_n(socket->rx_buf+rd_modulo, socket->recv_ptr        , thisgo1);
		wiz_read_n(socket->rx_buf          , socket->recv_ptr+thisgo1, thisgo2); 
		}
	if ( thisgo < available )
		socket->consumed += thisgo;
	else
		{
		socket->consumed = 0;
		wiz_write_word(sra+W5100_OFFSET_SOCKET_RX_RD0, rd+thisgo);
		wiz_write(sra+W5100_OFFSET_SOCKET_CR, W5100_SOCKET_CR_RECV); /* keep receiving */
		}
	socket->actual = thisgo;
	socket->state  = SS_ESTABLISHED;
	net_complete_op(socket);
	}

static void net_recv_more(SOCKET *socket)
	{
#ifdef DEBUG
    log_printf("net_recv_more: $");
    show_reg("socket = ", (word) socket);
#endif
	word sra = socket->sra;
	word available = wiz_read_word(sra+W5100_OFFSET_SOCKET_RX_RSR0) - socket->consumed;
	if ( available > 0 )
		net_recv_consume(socket, available);
	}

void net_recv(SOCKET *socket, byte *ptr, size_t length, NF nf, void *nfp)
	{
#ifdef DEBUG
    log_printf("net_recv: $");
    show_reg("sp = ", get_sp());
    show_reg("socket = ", (word) socket);
    show_reg("ptr = ", (word) ptr);
    show_reg("length = ", length);
#endif
	if ( socket->state != SS_ESTABLISHED )
		net_fatal(socket, "net_recv: socket ? not SS_ESTABLISHED\r\n$");
    if ( socket->fiber_op.queued ) fiber_dequeue (&socket->fiber_op);
    // net_fatal(socket, "net_recv: socket ? fiber already queued\r\n$");
#ifdef DEBUG
    log_printf ("net_recv: Socket $");
    log_uint ((socket->sra - W5100_ADDR_SOCKET) >> 8);
    log_printf (" callback $");
    log_word ((word) nf);
    log_printf ("\r\n$");
#endif
	socket->fiber_op.nf  = nf;
	socket->fiber_op.nfp = nfp;
	socket->recv_ptr     = ptr;
	socket->length       = length;
	socket->actual       = 0;
	if ( length == 0 )
		net_complete_op(socket);
	else
		{
		socket->state = SS_RECV;
		net_recv_more(socket);
		}
	}
/*...e*/

/*...snet_interrupt:0:*/

static void net_bad_interrupt (SOCKET *socket, byte ir, char *psState)
	{
    log_printf ("net_bad_interrupt: socket $");
    log_uint ((socket->sra - W5100_ADDR_SOCKET) >> 8);
    log_printf (" unexpected interrupt $");
    log_byte (ir);
    log_printf (" in $");
    net_fatal (socket, psState);
	}

static void net_interrupt(SOCKET *socket)
	{
	word sra = socket->sra;
	byte ir = wiz_read(sra+W5100_OFFSET_SOCKET_IR);
#ifdef DEBUG
    if (ir != 0)
        {
        log_printf ("net_interrupt: socket $");
        log_uint ((socket->sra - W5100_ADDR_SOCKET) >> 8);
        log_printf (" interrupt(s)");
        if ( ir & 0x01 ) log_printf (" CON$");
        if ( ir & 0x02 ) log_printf (" DISCON$");
        if ( ir & 0x04 ) log_printf (" RECV$");
        if ( ir & 0x08 ) log_printf (" TIMEOUT");
        if ( ir & 0x10 ) log_printf (" SEND_OK");
        log_printf (" in state$");
        log_printf (psStatus[socket->state]);
        }
#endif
//	wiz_write(sra+W5100_OFFSET_SOCKET_IR, ir); /* clear it */
	// socket->fiber_int_queued = FALSE;
//log_printf("s=$"); outbyte(socket->state); log_printf(" ir=$"); outbyte(ir); log_printf("\r\n$");
#ifdef DEBUG
    log_printf ("net_interrupt: socket $");
    log_uint ((socket->sra - W5100_ADDR_SOCKET) >> 8);
    log_printf (" interrupt $");
    log_byte (ir);
    log_printf (" in $");
#endif
	switch ( socket->state )
		{
		case SS_CONNECT:
#ifdef DEBUG
            log_printf ("SS_CONNECT\r\n$");
#endif
			if ( ir & W5100_SOCKET_IR_CON )
                {
                wiz_write(sra+W5100_OFFSET_SOCKET_IR, W5100_SOCKET_IR_CON); /* clear it */
				net_connect_ok(socket);
                }
			else if ( ir & (W5100_SOCKET_IR_DISCON|W5100_SOCKET_IR_TIMEOUT) )
                {
                wiz_write(sra+W5100_OFFSET_SOCKET_IR, ir); /* clear it */
				net_connect_error(socket);
                }
            else
                {
				net_bad_interrupt (socket, ir, "SS_CONNECT\r\n$");
                }
			break;
		case SS_LISTEN:
#ifdef DEBUG
            log_printf ("SS_LISTEN\r\n$");
#endif
			if ( ir & W5100_SOCKET_IR_CON )
                {
                wiz_write(sra+W5100_OFFSET_SOCKET_IR, W5100_SOCKET_IR_CON); /* clear it */
				net_listen_ok(socket);
                }
			else if ( ir & ~W5100_SOCKET_IR_RECV )
                {
				net_bad_interrupt (socket, ir, "SS_LISTEN\r\n$");
                }
			break;
		case SS_ESTABLISHED:
#ifdef DEBUG
            log_printf ("SS_ESTABLISHED\r\n$");
#endif
			if ( ir & ~(W5100_SOCKET_IR_RECV | W5100_SOCKET_IR_DISCON))
				net_bad_interrupt (socket, ir, "SS_ESTABLISHED\r\n$");
			break;
		case SS_SEND:
#ifdef DEBUG
            log_printf ("SS_SEND\r\n$");
#endif
			if ( ir & W5100_SOCKET_IR_SEND_OK )
                {
                wiz_write(sra+W5100_OFFSET_SOCKET_IR, W5100_SOCKET_IR_SEND_OK); /* clear it */
				net_send_chunk_ok(socket);
                }
			else if ( ir & (W5100_SOCKET_IR_DISCON|W5100_SOCKET_IR_TIMEOUT) )
                {
                wiz_write(sra+W5100_OFFSET_SOCKET_IR, ir); /* clear it */
				net_send_chunk_error(socket);
                }
			else if ( ir & ~W5100_SOCKET_IR_RECV )
                {
				net_bad_interrupt (socket, ir, "SS_SEND\r\n$");
                }
			break;
		case SS_RECV:
#ifdef DEBUG
            log_printf ("SS_RECV\r\n$");
#endif
			if ( ir & W5100_SOCKET_IR_RECV )
                {
                wiz_write(sra+W5100_OFFSET_SOCKET_IR, W5100_SOCKET_IR_RECV); /* clear it */
				net_recv_more(socket);
                }
			else if ( ir & (W5100_SOCKET_IR_DISCON|W5100_SOCKET_IR_TIMEOUT) )
                {
                wiz_write(sra+W5100_OFFSET_SOCKET_IR, ir); /* clear it */
				net_recv_error(socket);
                }
            else
                {
				net_bad_interrupt (socket, ir, "SS_RECV\r\n$");
                }
			break;
		case SS_ERROR:
#ifdef DEBUG
            log_printf ("SS_ERROR\r\n$");
#endif
            wiz_write(sra+W5100_OFFSET_SOCKET_IR, ir); /* clear it */
			break;
		case SS_DISCONNECT:
#ifdef DEBUG
            log_printf ("SS_DISCONNECT\r\n$");
#endif
			if ( ir & (W5100_SOCKET_IR_DISCON|W5100_SOCKET_IR_TIMEOUT) )
                {
                wiz_write(sra+W5100_OFFSET_SOCKET_IR, ir); /* clear it */
				net_disconnect_ok(socket);
                }
			break;
		default:
#ifdef DEBUG
            log_printf ("Unknown\r\n$");
#endif
			net_fatal(socket, "net_interrupt: socket ? unexpected interrupt\r\n$");
            break;
		}
	}
/*...e*/
/*...snet_handle_interrupts:0:*/
void net_handle_interrupts(void)
	{
	byte ir = wiz_read(W5100_ADDR_REG_IR);
	byte i;
#ifdef DEBUG
    if ( ir )
        {
        log_printf ("net_handle_interrupts: ir = $");
        log_byte (ir);
        log_printf ("\r\n$");
        }
#endif
	if ( ir & W5100_IR_CONFLICT )
		{
		log_printf("net_handle_interrupts: IP Conflict\r\n$");
		wiz_write(W5100_ADDR_REG_IR, W5100_IR_CONFLICT);
		}
	for ( i = 0; i < N_SOCKETS; i++ )
		{
		SOCKET *socket = &(net_sockets[i]);
		if ( ir & socket->int_bit )
			// if ( ! socket->fiber_int_queued )
			if ( ! socket->fiber_int.queued )
				{
                socket->fiber_int.nf     = (NF)net_interrupt;
                socket->fiber_int.nfp    = socket;
				// socket->fiber_int_queued = TRUE;
				fiber_enqueue(&(socket->fiber_int));
				}
		}
	}
/*...e*/

/*...snet_init:0:*/
void net_init(void)
	{
	byte i;
    net_config();
	list_init(&net_sockets_free);
	list_init(&net_sockets_waiters);
	for ( i = 0; i < N_SOCKETS; i++ )
		{
		SOCKET *socket = &(net_sockets[i]);
		socket->fiber_int.nf     = (NF)net_interrupt;
		socket->fiber_int.nfp    = socket;
        socket->fiber_int.queued = 0;
		// socket->fiber_int_queued = FALSE;
        socket->fiber_op.queued  = 0;
		socket->sra              = W5100_ADDR_SOCKET + (i<< 8);
		socket->tx_buf           = W5100_ADDR_TX     + (i<<11);
		socket->rx_buf           = W5100_ADDR_RX     + (i<<11);
		socket->int_bit          = (0x01<<i); /* per W5100_IR_Sx_INT */
		socket->state            = SS_FREE;
		list_add_last(&net_sockets_free, socket);
		}
	wiz_reset();
	wiz_write(W5100_ADDR_REG_IMR, 0x00); /* turn off all interrupts */
	wiz_write_n(W5100_ADDR_REG_SUBR0, ip_subnet, 4);
	wiz_write_n(W5100_ADDR_REG_SIPR0, ip, 4);
	wiz_write_n(W5100_ADDR_REG_GAR0 , ip_gateway, 4);
	wiz_write_n(W5100_ADDR_REG_SHAR0, mac, 6);
	wiz_write_word(W5100_ADDR_REG_RTR0, 2000);
	wiz_write(W5100_ADDR_REG_RCR, 10); /* retries */
	wiz_write(W5100_ADDR_REG_RMSR, 0x55); /* 2KB each for sockets 0,1,2,3 */
	wiz_write(W5100_ADDR_REG_TMSR, 0x55); /* 2KB each for sockets 0,1,2,3 */
	}
/*...e*/
